from setuptools import setup


setup(name='pll',
      version='0.1.1',
      author='Beta Faccion',
      author_email='betafcc@gmail.com',
      packages=['pll'],
      install_requires=['tqdm'],
      url='https://github.com/betafcc/pll',
      download_url='https://github.com/betafcc/pll/tarball/v0.1.1',
      )
